#include "chowdsp_StateVariableFilter.h"

namespace chowdsp
{
template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::StateVariableFilter()
{
    if constexpr (type == FilterType::MultiMode)
    {
        setMode(0);
    }
    setCutoffFrequency (static_cast<NumericType> (1000.0));
    setQValue (static_cast<NumericType> (1.0 / juce::MathConstants<double>::sqrt2));
    setGain (static_cast<NumericType> (1.0));
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
template <bool shouldUpdate>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::setCutoffFrequency (SampleType newCutoffFrequencyHz)
{
    if (newCutoffFrequencyHz >= sampleRate * 0.5)
    {
        DBG("WARNING: An attempt was made to set a filter cutoff frequency (" + juce::String(newCutoffFrequencyHz) + ") that is greater than the Nyquist frequency (" + juce::String(sampleRate * 0.5) + ").");
        newCutoffFrequencyHz = sampleRate * 0.5 - 1;
    }
    jassert (SIMDUtils::all (newCutoffFrequencyHz >= static_cast<NumericType> (0)));
    jassert (SIMDUtils::all (newCutoffFrequencyHz < static_cast<NumericType> (sampleRate * 0.5)));

    cutoffFrequency = newCutoffFrequencyHz;
    const auto w = juce::MathConstants<NumericType>::pi * cutoffFrequency / (NumericType) sampleRate;

    CHOWDSP_USING_XSIMD_STD (tan);
    g0 = tan (w);

    if constexpr (shouldUpdate)
        update();
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
template <bool shouldUpdate>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::setPeakFrequency (SampleType newFrequencyHz)
{
    jassert (SIMDUtils::all (newFrequencyHz >= static_cast<NumericType> (0)));
    jassert (SIMDUtils::all (newFrequencyHz < static_cast<NumericType> (sampleRate * 0.5)));

    cutoffFrequency = newFrequencyHz; // warning -- this is not the actual cutoff frequency! It's the peak frequency.
    const auto w = juce::MathConstants<NumericType>::pi * newFrequencyHz / (NumericType) sampleRate;

    CHOWDSP_USING_XSIMD_STD (tan);
    CHOWDSP_USING_XSIMD_STD (sqrt);
    const auto numerator = resonance * tan(w);
    const auto denominator = sqrt(resonance * resonance - 0.5);
    g0 = numerator / denominator;
    jassert (!std::isnan(g0));

    if constexpr (shouldUpdate)
        update();
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
template <bool shouldUpdate>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::setQValue (SampleType newResonance)
{
    jassert (SIMDUtils::all (newResonance > static_cast<NumericType> (0)));

    resonance = newResonance;
    k0 = (NumericType) 1.0 / resonance;
    k0A = k0 * A;

    if constexpr (shouldUpdate)
        update();
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
template <bool shouldUpdate>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::setGain (SampleType newGainLinear)
{
    jassert (SIMDUtils::all (newGainLinear > static_cast<NumericType> (0)));

    gain = newGainLinear;

    CHOWDSP_USING_XSIMD_STD (sqrt);
    A = sqrt (gain);
    sqrtA = sqrt (A);
    Asq = A * A;
    k0A = k0 * A;

    if constexpr (shouldUpdate)
        update();
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
template <bool shouldUpdate>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::setGainDecibels (SampleType newGainDecibels)
{
    setGain<shouldUpdate> (SIMDUtils::decibelsToGain (newGainDecibels));
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
template <StateVariableFilterType M>
std::enable_if_t<M == StateVariableFilterType::MultiMode, void>
    StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::setMode (NumericType newMode)
{
    this->mode = newMode;
    lowpassMult = (NumericType) 1 - (NumericType) 2 * juce::jmin ((NumericType) 0.5, newMode);
    bandpassMult = (NumericType) 1 - std::abs ((NumericType) 2 * (newMode - (NumericType) 0.5));
    highpassMult = (NumericType) 2 * juce::jmax ((NumericType) 0.5, newMode) - (NumericType) 1;

    //TODO restore original functionality

    // if constexpr(!unityGain)
    // {
    //     // use sin3db power law for mixing
    //     lowpassMult = std::sin (juce::MathConstants<NumericType>::halfPi * lowpassMult);
    //     bandpassMult = std::sin (juce::MathConstants<NumericType>::halfPi * bandpassMult);
    //     highpassMult = std::sin (juce::MathConstants<NumericType>::halfPi * highpassMult);
    //
    //     // the BPF is a little bit quieter by design, so let's compensate here for a smooth transition
    //     bandpassMult *= juce::MathConstants<NumericType>::sqrt2;
    // }

}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
template <StateVariableFilterType M>
std::enable_if_t<M == StateVariableFilterType::MultiMode, bool>
    StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::updateParameters (SampleType newFrequency, SampleType newResonance, NumericType newMode, bool peak)
{
    bool flag = false;
    if(newMode != mode)
    {
        setMode (newMode);
        flag = true;
    }
    if(peak)
    {
        bool resonanceChanged = false;
        if (newResonance != resonance)
        {
            setQValue<false> (newResonance);
            resonanceChanged = true;
            flag = true;
        }
        if (newFrequency != cutoffFrequency || resonanceChanged)
        {
            setPeakFrequency<false> (newFrequency);
            flag = true;
        }
    }
    else
    {
        if (newResonance != resonance)
        {
            setQValue<false> (newResonance);
            flag = true;
        }

        if (newFrequency != cutoffFrequency)
        {
            setCutoffFrequency<false> (newFrequency);
            flag = true;
        }
    }

    if(flag)
    {
        update();
    }

    return flag;
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::prepare (const juce::dsp::ProcessSpec& spec)
{
    jassert (spec.sampleRate > 0);
    jassert (spec.numChannels > 0);

    sampleRate = spec.sampleRate;

    if constexpr (maxChannelCount == dynamicChannelCount)
    {
        ic1eq.resize (spec.numChannels);
        ic2eq.resize (spec.numChannels);
    }
    else
    {
        jassert (spec.numChannels <= maxChannelCount);
    }

    reset();
    DBG("Init cutoff frequency: " + juce::String(cutoffFrequency));
    setCutoffFrequency (cutoffFrequency);
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::reset()
{
    for (auto v : { &ic1eq, &ic2eq })
        std::fill (v->begin(), v->end(), static_cast<SampleType> (0));
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::snapToZero() noexcept // NOSONAR (cannot be const)
{
#if JUCE_SNAP_TO_ZERO
    for (auto v : { &ic1eq, &ic2eq })
        for (auto& element : *v)
            juce::dsp::util::snapToZero (element);
#endif
}

template <typename SampleType, StateVariableFilterType type, size_t maxChannelCount, bool unityGain>
void StateVariableFilter<SampleType, type, maxChannelCount, unityGain>::update()
{
    SampleType g, k;
    if constexpr (type == FilterType::Bell)
    {
        g = g0;
        k = k0 / A;
    }
    else if constexpr (type == FilterType::LowShelf)
    {
        g = g0 / sqrtA;
        k = k0;
    }
    else if constexpr (type == FilterType::HighShelf)
    {
        g = g0 * sqrtA;
        k = k0;
    }
    else
    {
        g = g0;
        k = k0;
    }

    const auto gk = g + k;
    a1 = (NumericType) 1.0f / ((NumericType) 1.0f + g * gk);
    a2 = g * a1;
    a3 = g * a2;
    ak = gk * a1;

    if constexpr(unityGain)
    {
        oneOverPeakGain = 1.0f / getPeakGain();
    }
}
} // namespace chowdsp
